import json

from omnomnom import exceptions
from omnomnom.models import Message


class BasicMarshaller(object):

    def prepare_request(self, msg):
        # Given omnomnom.models.Message build request
        # that's send to service
        raise NotImplementedError

    def process_request_data(self, data):
        # Given raw request data send to socket
        # returns omnomnom.models.Message
        raise NotImplementedError

    def prepare_response(self, result=None, err=None):
        # Given executable result or error
        # prepares response that's send to client
        raise NotImplementedError

    def process_response_data(self, data):
        # Given raw response data build data structure
        # that's usable for client
        raise NotImplementedError


class LoadDumpMarshaller(BasicMarshaller):
    executable_field = 'executable'

    def dumps(self, data):
        raise NotImplementedError

    def loads(self, data):
        raise NotImplementedError

    def prepare_request(self, msg):
        request = {
            'executable': msg.executable,
            'args': msg.args,
            'kwargs': msg.kwargs,
        }
        return self.dumps(request)

    def process_request_data(self, data):
        data = self.loads(data)

        try:
            executable = data[self.executable_field]
        except KeyError:
            msg = u"No executable defined in message"
            raise exceptions.DemarshallingExeption(msg)

        args = data.get('args', ())
        kwargs = data.get('kwargs', {})
        return Message(
            executable=executable,
            args=args, kwargs=kwargs
        )

    def prepare_response(self, result=None, err=None):
        if err and not isinstance(err, (str, unicode)):
            err = str(err)
        response = {
            'result': result,
            'error': err
        }
        return self.dumps(response)

    def process_response_data(self, data):
        return self.loads(data)


class JsonMarshaller(LoadDumpMarshaller):

    def dumps(self, data):
        return json.dumps(data).encode('utf-8')

    def loads(self, data):
        return json.loads(data.decode('utf-8'))


class BsonMarshaller(LoadDumpMarshaller):

    def dumps(self, data):
        import bson
        return bson.dumps(data)

    def loads(self, data):
        import bson
        return bson.loads(data)


class MsgPackMarshaller(LoadDumpMarshaller):

    def dumps(self, data):
        import msgpack
        return msgpack.packb(data, use_bin_type=True)

    def loads(self, data):
        import msgpack
        return msgpack.unpackb(data, encoding='utf-8')


import random

import nanomsg

from omnomnom import models
from omnomnom import exceptions


class LycheeClient(object):
    default_send_timeout = 100

    def __init__(self, addresses, standalone=False, **kwargs):
        self.addresses = [addresses] if isinstance(addresses, (str, unicode))\
            else addresses
        self.standalone = standalone
        self.propagate_exc = kwargs.pop('propagate_exc', False)

    def __enter__(self):
        self.socket = models.Pip(nanomsg.PUSH)
        if self.standalone:
            address = self.get_worker_address()
            self.socket.connect(address=address)
        else:
            self.socket.connect(address=self.addresses[0])
        self.socket._set_send_timeout(self.default_send_timeout)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.socket.close()

    def get_worker_address(self):
        # when there are multiple worker families
        # (e.g few workers bind to one socket and few bind to another one)
        # this function picks one of them and binds to particular socket
        return random.choice(self.addresses)

    def send_task(self, task_instance):
        transport_data = task_instance.to_json()
        self.socket.send_json(transport_data)

    def execute(self, fun, args=None, kwargs=None):
        args = args if args else ()
        kwargs = kwargs if kwargs else {}
        task = fun.__name__ if not isinstance(fun, (str, unicode)) else fun

        is_from_main = lambda x: x.__module__ == '__main__'

        if not isinstance(fun, (str, unicode)) and is_from_main(fun):
            raise ValueError(u'Cannot enqueue functions from __main__ module')

        try:
            t = models.Task(
                task_name=task,
                task_args=args, task_kwargs=kwargs
            )
            self.send_task(t)
        except nanomsg.NanoMsgAPIError:
            if self.propagate_exc:
                msg = u"Timeout {0} exceeded".format(self.default_send_timeout)
                raise exceptions.PipConnectionException(msg)



import time
import pickle
import logging
from threading import Thread
from datetime import datetime, timedelta

from redis.exceptions import ResponseError


class BaseResultBackend(object):

    def add_task(self, task):
        raise NotImplementedError

    def delete_task(self, task):
        raise NotImplementedError

    def update_task(self, task):
        raise NotImplementedError

    def get_task(self, task_id):
        raise NotImplementedError

    def cleanup(self, task_result_expires):
        raise NotImplementedError

    def get_tasks(self, **kwargs):
        raise NotImplementedError

    def setup(self, **kwargs):
        logging.info(u"Setting up result backend")

    def terminate(self):
        logging.info(u"Terminating result backend")


class DummyResultBackend(BaseResultBackend):

    def add_task(self, task):
        pass

    def delete_task(self, task):
        pass

    def update_task(self, task):
        pass

    def get_task(self, task_id):
        pass

    def cleanup(self, task_result_expires):
        pass

    def get_tasks(self, **kwargs):
        return []

    def setup(self, **kwargs):
        super(DummyResultBackend, self).setup(**kwargs)

    def terminate(self):
        super(DummyResultBackend, self).terminate()


class RedisResultBackend(BaseResultBackend):
    task_prefix = 'omnomnom_task_'
    # by default task expire after 1 day
    expires_after = 3600 * 24

    def __init__(self, host='localhost', port=6379, db=0):
        import redis
        self.r = redis.StrictRedis(host=host, port=port, db=db)

    def add_task(self, task):
        self.r.set(
            '{0}{1}'.format(self.task_prefix, task.task_id),
            pickle.dumps(task)
        )

    def delete_task(self, task):
        self.r.delete(
            '{0}{1}'.format(self.task_prefix, task.task_id)
        )

    def update_task(self, task):
        self.r.set(
            '{0}{1}'.format(self.task_prefix, task.task_id),
            pickle.dumps(task)
        )

    def get_task(self, task_id):
        return pickle.loads(
            self.r.get('{0}{1}'.format(self.task_prefix, task_id))
        )

    def get_tasks(self, **kwargs):
        match = kwargs.pop('match', '{0}*'.format(self.task_prefix))

        try:
            for key in self.r.scan_iter(match=match):
                yield (key, pickle.loads(self.r.get(key)))
        except ResponseError:
            for key in self.r.keys(pattern=match):
                yield (key, pickle.loads(self.r.get(key)))

    def cleanup(self, task_result_expires):
        now = datetime.now()
        cleaned_tasks = 0

        for key in self.r.keys('{0}*'.format(self.task_prefix)):
            task = self.get_task(
                key.replace('{0}'.format(self.task_prefix), '')
            )
            if task.has_finished:
                if task.completed_at < (now - timedelta(seconds=task_result_expires)):
                    cleaned_tasks += 1
                    self.delete_task(task)

        return cleaned_tasks

    def setup(self, **kwargs):
        super(RedisResultBackend, self).setup(**kwargs)
        self.monitor_thread_signal = True
        self.monitor_thread = Thread(target=self.monitor_tasks)
        self.monitor_thread.start()

    def monitor_tasks(self):
        elapsed, step = 0, 1

        while self.monitor_thread_signal:
            if self.expires_after <= elapsed:
                cleaned = self.cleanup(self.expires_after)
                logging.debug(u"Results Cleanup: {0} tasks.".format(cleaned))
                elapsed = 0
            time.sleep(1)
            elapsed += step

    def terminate(self):
        super(RedisResultBackend, self).terminate()
        self.monitor_thread_signal = False


from omnomnom.generics.taskqueue import client


def task():
    import time
    print "executing task"
    time.sleep(5)
    print "Done executuing"
    return "sth"


workers = [
    u'tcp://127.0.0.1:7105',
    u'tcp://127.0.0.1:7106'
]

with client.LycheeClient(workers, standalone=True, propagate_exc=True) as c:
    c.execute('examples.taskqueue.multi.task')


import sys
import logging

import nanomsg

from omnomnom import handlers
from omnomnom import models
from omnomnom import result_backends
from omnomnom import utils


class LycheeWorker(object):
    process_fun = 'recv_json'

    def __init__(self, address, **kwargs):
        self.address = address
        self.name = self.get_name(**kwargs)
        self.standalone = self.is_standalone(**kwargs)
        self.socket = self.get_socket(**kwargs)
        self.setup_socket(**kwargs)

        self.result_backend = self.get_result_backend(**kwargs)
        self.init_result_backend(**kwargs)
        self.handler = self.get_handler(**kwargs)
        self.log = logging.getLogger(__name__)

    def get_name(self, **kwargs):
        name = kwargs.get('name', None)
        return name if name else u'W-{0}'.format(utils.generate_name())

    def get_socket(self, **kwargs):
        socket = models.Pip(nanomsg.PULL)
        return socket

    def get_result_backend(self, **kwargs):
        result_backend = kwargs.get('result_backend', None)
        return result_backend or result_backends.DummyResultBackend()

    def init_result_backend(self, **kwargs):
        self.result_backend.setup(**kwargs)

    def is_standalone(self, **kwargs):
        return kwargs.get('standalone', False)

    def setup_socket(self, **kwargs):
        if self.standalone:
            self.socket.bind(self.address)
        else:
            self.socket.connect(self.address)

    def get_handler(self, **kwargs):
        if self.standalone:
            return handlers.FuturesHandler(
                result_backend=self.result_backend, **kwargs
            )
        return handlers.BasicHandler(result_backend=self.result_backend, **kwargs)

    def stop(self):
        self.log.info(u"Stopping {0} gracefully.".format(self.name))
        self.handler.close()
        self.socket.close()
        self.result_backend.terminate()
        sys.exit(1)

    def start(self):
        self.log.info(u"{0} Started on {1}".format(self.name, self.address))

        if self.standalone:
            self.log.info(u"{0} runs in standalone mode".format(self.name))

        while True:
            try:
                data = getattr(self.socket, self.process_fun)()
                task = models.Task.from_task_data(data)
                self.log.debug(u'{0} received {1}'.format(
                    self.name, task.get_info()
                ))
                self.handler.process_task(task)
            except Exception as e:
                self.log.error(e, exc_info=1)

        self.socket.close()


import signal
import argparse

from omnomnom import utils
from omnomnom.generics.taskqueue import LycheeWorker


parser = argparse.ArgumentParser(description='Omnomnom Worker Runner')

parser.add_argument(
    '--address', dest='address', type=str,
    help='Socket to connect to', default='tcp://127.0.0.1:7105'
)
parser.add_argument(
    '--name', dest='name', type=str,
    help='Worker name'
)
parser.add_argument(
    '--log-level', dest='log_level', type=str,
    default='INFO'
)
parser.add_argument(
    '--standalone', dest='standalone', action='store_true',
    help='Enables to run worker in standalone mode (without queue as a proxy)'
)
parser.add_argument(
    '--max-workers', dest='max_workers', type=int,
    help='Number of workers to define when runnning in standalone mode',
    default=4
)


def shutdown(worker_instance):
    worker_instance.stop()


def get_options():
    arguments = parser.parse_args()
    return arguments


def main():
    args = get_options()
    utils.setup_loghandlers(args.log_level)

    worker = LycheeWorker(
        address=args.address, name=args.name,
        standalone=args.standalone, max_workers=args.max_workers
    )

    shutdown_handler = lambda sig, frame: shutdown(worker)
    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)

    worker.start()


import uuid
import json
from datetime import datetime

from nanomsg import Socket


class Pip(Socket):

    def send_json(self, msg, flags=0, **kwargs):
        msg = json.dumps(msg, **kwargs).encode('utf8')
        self.send(msg, flags)

    def recv_json(self, buf=None, flags=0):
        msg = self.recv(buf, flags)
        return json.loads(msg)


class Task(object):

    class Status:
        SENT = 'sent'
        PROCESSING = 'processing'
        FINISHED = 'finished'
        FAILED = 'failed'

    def __init__(self, *args, **kwargs):
        self.task_name = kwargs.get('task_name')
        self.task_args = kwargs.get('task_args', None) or ()
        self.task_kwargs = kwargs.get('task_kwargs', None) or {}
        self.created_at = kwargs.get('created_at', datetime.now())
        self.completed_at = kwargs.pop('completed_at', None)
        self.failed_at = kwargs.pop('failed_at', None)
        self.task_id = kwargs.get('task_id', uuid.uuid4().hex)
        self.status = kwargs.get('status', Task.Status.SENT)
        self.result = kwargs.get('result', None)
        self.failure_reason = kwargs.pop('failure_reason', None)

    @classmethod
    def from_task_data(cls, data):
        return Task(**data)

    @property
    def has_finished(self):
        status_list = [Task.Status.FINISHED]
        return self.status in status_list

    def finalize(self):
        self.status = Task.Status.FINISHED
        self.completed_at = datetime.now()

    def failure(self, reason):
        self.status = Task.Status.FAILED
        self.failed_at = datetime.now()
        self.failure_reason = reason

    def get_info(self):
        return {
            'id': self.task_id,
            'name': self.task_name,
            'args': self.task_args,
            'kwargs': self.task_kwargs
        }

    def to_json(self):
        return {
            'task_id': self.task_id,
            'task_name': self.task_name,
            'task_args': self.task_args,
            'task_kwargs': self.task_kwargs,
            'created_at': str(self.created_at),
            'status': self.status
        }


class Executable(object):

    def __call__(self, *args, **kwargs):
        # Executable receives service instance in kwargs
        pass


class Message(object):

    def __init__(self, executable, args, kwargs, *a, **kw):
        self.executable = executable
        self.args = args if args else ()
        self.kwargs = kwargs if kwargs else {}



import sys
import logging

import nanomsg

from omnomnom import models


logger = logging.getLogger(__name__)


class LycheeQueue(object):

    def __init__(self, *args, **kwargs):
        self.proxied = 0
        self.push_address = kwargs.pop('push_address')
        self.pull_address = kwargs.pop('pull_address')
        self.push_socket = kwargs.pop('push_socket', self.get_push_socket())
        self.pull_socket = kwargs.pop('pull_socket', self.get_pull_socket())
        self.push_socket.bind(self.push_address)
        self.pull_socket.bind(self.pull_address)
        self.log = logger

    def get_push_socket(self):
        socket = models.Pip(nanomsg.PUSH)
        return socket

    def get_pull_socket(self):
        socket = models.Pip(nanomsg.PULL)
        return socket

    def stop(self):
        self.log.info(u"Stopping Queue gracefully")
        self.push_socket.close()
        self.pull_socket.close()
        sys.exit(1)

    def start(self):
        self.log.info(u"Queue instance started. Frontend ({0}) <-- > Backend ({1})".format(
            self.pull_address, self.push_address
        ))

        pull_fun = 'recv'
        push_fun = 'send'

        while True:
            try:
                data = getattr(self.pull_socket, pull_fun)()
                self.log.info(u"Proxying message")
                self.log.debug(data)
                self.proxied += 1
                getattr(self.push_socket, push_fun)(data)
                self.log.info(u"Proxied {0} messages".format(self.proxied))
            except Exception as e:
                self.log.error(e, exc_info=1)

        self.pull_socket.close()
        self.push_socket.close()



import random

from omnomnom.generics.micro import PubFruit


if __name__ == '__main__':
    address = 'tcp://127.0.0.1:5005'
    service = PubFruit(address=address, log_level='INFO')

    while True:
        response = service.marshaller.prepare_response(
            result=random.randint(1, 10000)
        )
        service.send_response(response)


import random

from omnomnom.generics.micro import MicroClient


if __name__ == '__main__':
    with MicroClient(address='tcp://127.0.0.1:5001') as client:
        print client.get(executable='ping')
        print client.get(executable='pong')
        print client.get(executable='exc')
        print client.get(executable='heavy')
        print client.get(executable='long', kwargs={
            'delta': random.randint(1, 5)
        })


import signal
import argparse

from omnomnom import utils
from omnomnom.generics.taskqueue import LycheeQueue


parser = argparse.ArgumentParser(description='Omnomnom Queue Runner')
parser.add_argument(
    '--frontend', dest='frontend', type=str,
    help='Frontend socket address', default='tcp://127.0.0.1:7104'
)
parser.add_argument(
    '--backend', dest='backend', type=str,
    help='Backend socket address', default='tcp://127.0.0.1:7105'
)
parser.add_argument(
    '--log-level', dest='log_level', type=str,
    default='INFO'
)


def shutdown(queue_instance):
    queue_instance.stop()


def get_options():
    arguments = parser.parse_args()
    return arguments


def main():
    args = get_options()
    utils.setup_loghandlers(args.log_level)

    queue = LycheeQueue(
        pull_address=args.frontend,
        push_address=args.backend
    )

    shutdown_handler = lambda sig, frame: shutdown(queue)
    signal.signal(signal.SIGINT, shutdown_handler)
    signal.signal(signal.SIGTERM, shutdown_handler)

    queue.start()


class OmnomnomException(Exception):
    """
    Base Omnomnom Exception Class.
    """


class PipConnectionException(OmnomnomException):
    """
    Raised when nanomsg cannot handle socket communication.
    """


class MarshallingException(OmnomnomException):
    """
    Raised during marshalling process.
    """


class DemarshallingExeption(OmnomnomException):
    """
    Raised during demarshalling process.
    """


class MicroServiceInitializationException(OmnomnomException):
    """
    Raised when service cannot be properly initialized
    based on its configuration.
    """


class MicroServiceRegistrationException(OmnomnomException):
    """
    Raised when executable cannot be registered.
    """


class MicroServiceExecutionException(OmnomnomException):
    """
    Raised when executable cannot be executed.
    or is not registered.
    """


class MicroServiceRunnerException(OmnomnomException):
    """
    Raised when service cannot be imported.
    """


import logging

from omnomnom.models import Executable
from omnomnom.generics.micro import Fruit


class PingExecutable(Executable):

    def __call__(self, *args, **kwargs):
        return 'pong'


class PongExecutable(Executable):

    def __call__(self, *args, **kwargs):
        return 'ping'


class ExceptionExecutable(Executable):

    def __call__(self, *args, **kwargs):
        raise Exception(u"Kaboom")


class LongExecutable(Executable):

    def __call__(self, *args, **kwargs):
        import time
        delta = int(kwargs['delta'])
        logging.info(u"Sleeping for {0} second(s)".format(delta))
        time.sleep(delta)
        return 'done'


class HeavyExecutable(Executable):

    def __call__(self, *args, **kwargs):
        import urllib2
        data = urllib2.urlopen('http://www.google.pl').read()
        # replaces invalid utf-8 characters
        return data.decode('utf-8', 'replace')


class PingPongService(Fruit):
    ping = PingExecutable()
    pong = PongExecutable()
    long = LongExecutable()
    heavy = HeavyExecutable()
    exc = ExceptionExecutable()


if __name__ == '__main__':
    service = PingPongService(address='tcp://127.0.0.1:5001', log_level='DEBUG')
    service.run()


# Inspired by: https://github.com/walkr/nanoservice/tree/master/benchmarks

import time
from multiprocessing import Process

import nanomsg
from omnomnom.models import Executable
from omnomnom.generics.micro import Fruit


class MultiplyExecutable(Executable):

    def __call__(self, x, y, **kwargs):
        return x + y


class JsonTestService(Fruit):
    multiply = MultiplyExecutable()


class JsonPubTestService(Fruit):
    socket_type = nanomsg.PUB
    multiply = MultiplyExecutable()


class Benchmark(object):

    def __init__(self, name, service_class, message_num, address, task, client_class):
        self.name = name
        self.service_class = service_class
        self.client_class = client_class
        self.message_num = message_num
        self.address = address
        self.task = task

    def handle_service(self, service):
        service.handle()

    def handle_client(self, client, task, args):
        res = client.get(task, args=args)
        assert not res['error']

    def run(self):
        print('')
        print(self.name)
        print('Dealing with {0} messages'.format(self.message_num))
        print('-----------------------------')

        service_process = Process(
            target=self.start_service, args=(self.address, self.message_num)
        )
        service_process.start()
        time.sleep(0.1)

        with self.client_class(address=self.address) as client:
            self.execute_benchmark(client, self.message_num, self.task)

        time.sleep(0.2)
        service_process.terminate()

    def execute_benchmark(self, client, message_num, task):

        pairs = [
            (x, x+1) for x in range(message_num)
        ]

        started = time.time()
        for pair in pairs:
            self.handle_client(client, task, pair)
        duration = time.time() - started
        print('Client stats:')
        self.print_stats(message_num, duration)

    def start_service(self, address, n):
        service = self.service_class(address=address)
        started = time.time()

        for _ in range(n):
            self.handle_service(service)

        duration = time.time() - started

        time.sleep(0.1)
        print('Service stats:')
        self.print_stats(n, duration)
        return

    def print_stats(self, n, duration):
        pairs = [
            ('Total messages', n),
            ('Total duration (s)', duration),
            ('Throughput (msg/s)', n/duration)
        ]
        for pair in pairs:
            label, value = pair
            print(' * {:<25}: {:10,.2f}'.format(label, value))


class RawBenchmark(Benchmark):

    def handle_service(self, service):
        data = service.socket.recv()
        service.socket.send(data)

    def handle_client(self, client, task, args):
        request = 'data'
        client.socket.send(request)
        response = client.socket.recv()
        assert request == response


from omnomnom.generics.taskqueue import client


def task():
    import time
    print "executing task"
    time.sleep(5)
    print "Done executuing"
    return "sth"


with client.LycheeClient(u'tcp://127.0.0.1:7104', propagate_exc=True) as c:
    c.execute('examples.taskqueue.proxy.task')


import pprint
import logging
import functools

import nanomsg

from omnomnom import models
from omnomnom import exceptions
from omnomnom.generics.micro import utils
from omnomnom.utils import setup_loghandlers
from omnomnom.generics.micro import marshalling


class MicroServiceExecutionMeta(type):

    def __new__(cls, name, bases, attrs):
        new_attrs = attrs.copy()
        new_attrs['executables'] = {}

        for (el, val) in attrs.iteritems():
            if isinstance(val, models.Executable):
                is_valid, error = utils.validate_executable(val)
                if not is_valid:
                    raise exceptions.MicroServiceInitializationException(error)

                executable_name = getattr(val, 'name', el)
                new_attrs['executables'][executable_name] = val

        return super(MicroServiceExecutionMeta, cls).__new__(cls, name, bases, new_attrs)


class BasicMicroService(object):
    __metaclass__ = MicroServiceExecutionMeta

    name = None
    authentication_class = None
    socket_type = nanomsg.REP
    marshaller = marshalling.JsonMarshaller()

    def __init__(self, address, **kwargs):
        self.address = address
        self.name = self.get_name(**kwargs)
        self.marshaller = self.get_marshaller(**kwargs)
        self.authentication_class = self.get_authentication_class(**kwargs)
        self.socket_type = self.get_socket_type(**kwargs)
        self.socket = kwargs.pop('socket', self.get_socket(**kwargs))
        self.setup_socket(**kwargs)
        self.init_service(**kwargs)

    def get_marshaller(self, **kwargs):
        return self.marshaller

    def get_authentication_class(self, **kwargs):
        return self.authentication_class

    def get_name(self, **kwargs):
        return self.name or self.__class__.__name__

    def get_socket_type(self, **kwargs):
        return self.socket_type

    def get_socket(self, **kwargs):
        if not self.socket_type:
            raise RuntimeError(u"socket_type must be specified")

        socket = models.Pip(self.socket_type)
        return socket

    def setup_socket(self, **kwargs):
        # used to add initial socket options
        # and bind/connect
        self.bind()

    def receive_data(self):
        return self.socket.recv()

    def send_response(self, response):
        self.socket.send(response)

    def connect(self):
        self.socket.connect(self.address)

    def bind(self):
        self.socket.bind(self.address)

    def run(self, *args, **kwargs):
        raise NotImplementedError

    def init_service(self, **kwargs):
        # log handlers and additional setup
        # like for instance database connection
        log_level = kwargs.get('log_level', 'DEBUG')
        setup_loghandlers(level=log_level)

    def register_executable(self, name, executable):
        if name in self.executables:
            msg = u"{0} is already registered".format(name)
            raise exceptions.MicroServiceRegistrationException(msg)

        is_valid, error = utils.validate_executable(executable)

        if not is_valid:
            raise exceptions.MicroServiceRegistrationException(error)

        self.executables[name] = executable

    def to_registry(self, name, *args, **kwargs):
        return functools.partial(
            utils.register_executable, registry=self.executables,
            name=name, *args, **kwargs
        )

    def get_executables_information(self):
        printer = pprint.PrettyPrinter(indent=4)
        printer.pprint(self.executables)

    def execute(self, msg):
        try:
            executable = self.executables[msg.executable]
        except KeyError:
            msg = u"Executable {0} is not registered".format(msg.executable)
            raise exceptions.MicroServiceExecutionException(msg)

        return executable(service=self, *msg.args, **msg.kwargs)


class Fruit(BasicMicroService):

    def handle(self):
        request_data = self.receive_data()
        #logging.debug(u"Service received {0}".format(request_data))

        try:
            msg = self.marshaller.process_request_data(data=request_data)
            result = self.execute(msg)
            response = self.marshaller.prepare_response(result=result)
            self.send_response(response)
        except Exception as e:
            # import traceback
            # traceback.print_exc()
            logging.error(e)
            response = self.marshaller.prepare_response(err=e)
            self.send_response(response)

    def run(self, *args, **kwargs):
        logging.info(u"Starting {0}".format(self.name))
        logging.info(u"Registered executables: {0}".format(
            self.executables.keys())
        )

        while True:
            self.handle()


class PubFruit(BasicMicroService):
    socket_type = nanomsg.PUB


from omnomnom.models import Executable
from omnomnom.generics.micro import marshalling
from omnomnom.generics.micro import Fruit
from omnomnom.generics.micro import MicroClient

import base


class MultiplyExecutable(Executable):

    def __call__(self, x, y, **kwargs):
        return x + y


class JsonTestService(Fruit):
    marshaller = marshalling.JsonMarshaller()
    multiply = MultiplyExecutable()


class BsonTestService(Fruit):
    marshaller = marshalling.BsonMarshaller()
    multiply = MultiplyExecutable()


class MsgPackTestService(Fruit):
    marshaller = marshalling.MsgPackMarshaller()
    multiply = MultiplyExecutable()


class JsonClient(MicroClient):
    marshaller = marshalling.JsonMarshaller()


class BsonClient(MicroClient):
    marshaller = marshalling.BsonMarshaller()


class MsgPackClient(MicroClient):
    marshaller = marshalling.MsgPackMarshaller()


if __name__ == '__main__':
    message_num = 20000

    benchmark = base.Benchmark(
        name='Json Marshaller Req-Rep over IPC', service_class=JsonTestService,
        message_num=message_num, address='ipc:///tmp/bench-req-rep-ipc.sock',
        task='multiply', client_class=JsonClient
    )
    benchmark.run()

    benchmark = base.Benchmark(
        name='Json Marshaller Req-Rep over TCP', service_class=JsonTestService,
        message_num=message_num, address='tcp://127.0.0.1:5051',
        task='multiply', client_class=JsonClient
    )
    benchmark.run()

    benchmark = base.Benchmark(
        name='Bson Marshaller Req-Rep over IPC', service_class=BsonTestService,
        message_num=message_num, address='ipc:///tmp/bench-req-rep-ipc.sock',
        task='multiply', client_class=BsonClient
    )
    benchmark.run()

    benchmark = base.Benchmark(
        name='Bson Marshaller Req-Rep over TCP', service_class=BsonTestService,
        message_num=message_num, address='tcp://127.0.0.1:5051',
        task='multiply', client_class=BsonClient
    )
    benchmark.run()

    benchmark = base.Benchmark(
        name='Msgpack Marshaller Req-Rep over IPC', service_class=MsgPackTestService,
        message_num=message_num, address='ipc:///tmp/bench-req-rep-ipc.sock',
        task='multiply', client_class=MsgPackClient
    )
    benchmark.run()

    benchmark = base.Benchmark(
        name='Msgpack Marshaller Req-Rep over TCP', service_class=MsgPackTestService,
        message_num=message_num, address='tcp://127.0.0.1:5051',
        task='multiply', client_class=MsgPackClient
    )
    benchmark.run()

    message_num = 50000

    benchmark = base.RawBenchmark(
        name='RAW Req-Rep over IPC', service_class=JsonTestService,
        message_num=message_num, address='ipc:///tmp/bench-req-rep-ipc.sock',
        task='multiply', client_class=MicroClient
    )
    benchmark.run()

    benchmark = base.RawBenchmark(
        name='RAW Req-Rep over TCP', service_class=JsonTestService,
        message_num=message_num, address='tcp://127.0.0.1:5051',
        task='multiply', client_class=MicroClient
    )
    benchmark.run()




import logging

from concurrent.futures import ThreadPoolExecutor

from omnomnom import models
from omnomnom import utils


class BasicHandler(object):

    def __init__(self, result_backend, *args, **kwargs):
        self.result_backend = result_backend

    def run_task(self, task, func, args, kwargs):
        result = func(*args, **kwargs)
        task.result = result
        task.finalize()
        self.result_backend.update_task(task)

    def handle_failed(self, task, e):
        task.failure(reason=str(e))
        self.result_backend.update_task(task)

    def process_task(self, task):
        try:
            func = utils.get_func(task.task_name)
            task.status = models.Task.Status.PROCESSING
            self.result_backend.add_task(task)
            self.run_task(
                task=task, func=func,
                args=task.task_args, kwargs=task.task_kwargs
            )
        except Exception as e:
            logging.error(e)
            self.handle_failed(task=task, e=e)

    def close(self):
        pass


class FuturesHandler(BasicHandler):
    DEFAULT_WORKER_NUM = 4
    WAIT_FOR_RESOURCES = False

    def __init__(self, *args, **kwargs):
        super(FuturesHandler, self).__init__(*args, **kwargs)
        max_workers = kwargs.get('max_workers', self.DEFAULT_WORKER_NUM)
        self.executor = ThreadPoolExecutor(max_workers=max_workers)
        msg = (
            u"Initialized Futures Handler "
            u"that uses ThreadPoolExecutor with {0} workers"
        )
        logging.debug(msg.format(max_workers))

    def handle_failed(self, task, e):
        pass

    def _update_task(self, future, task):
        e = future.exception()
        if not e:
            task.result = future.result()
            task.finalize()
            self.result_backend.update_task(task)
        else:
            task.failure(reason=str(e))
            self.result_backend.update_task(task)

    def run_task(self, task, func, args, kwargs):
        future = self.executor.submit(
            func, *args, **kwargs
        )
        future.add_done_callback(lambda f: self._update_task(f, task))

    def close(self):
        self.executor.shutdown(wait=self.WAIT_FOR_RESOURCES)


import nanomsg

from omnomnom import models
from omnomnom import exceptions
from omnomnom.generics.micro import marshalling


class BasicMicroClient(object):
    socket_type = nanomsg.REQ
    default_send_timeout = 100
    marshaller = marshalling.JsonMarshaller()

    def __init__(self, address, **kwargs):
        self.address = address
        self.propagate_exc = kwargs.pop('propagate_exc', False)

    def get_socket(self):
        return models.Pip(self.socket_type)

    def __enter__(self):
        self.socket = self.get_socket()
        self.socket.connect(self.address)
        self.socket._set_send_timeout(self.default_send_timeout)
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        self.socket.close()


class MicroClient(BasicMicroClient):

    def send_request(self, request):
        self.socket.send(request)

    def get(self, executable, args=None, kwargs=None):
        msg = models.Message(
            executable=executable,
            args=args, kwargs=kwargs
        )
        request = self.marshaller.prepare_request(msg)

        try:
            self.send_request(request=request)
            response_data = self.socket.recv()
            return self.marshaller.process_response_data(
                data=response_data
            )
        except nanomsg.NanoMsgAPIError:
            if self.propagate_exc:
                msg = u"Timeout {0} exceeded".format(self.default_send_timeout)
                raise exceptions.PipConnectionException(msg)


class MicroSubClient(BasicMicroClient):
    socket_type = nanomsg.SUB

    def __init__(self, address, **kwargs):
        super(MicroSubClient, self).__init__(address, **kwargs)
        self.subscribe_to = kwargs.pop('subscribe_to', '')

    def get_socket(self):
        socket = models.Pip(self.socket_type)
        socket.set_string_option(
            nanomsg.SUB, nanomsg.SUB_SUBSCRIBE, self.subscribe_to
        )
        return socket

    def receive(self):
        response_data = self.socket.recv()
        return self.marshaller.process_response_data(
            data=response_data
        )

    def get_responses(self):
        while True:
            yield self.receive()



import os
import sys
import inspect
import argparse

from omnomnom import exceptions
from omnomnom.generics.micro.service import BasicMicroService


parser = argparse.ArgumentParser(description='Omnomnom Service Runner')
parser.add_argument('service', nargs='?')
parser.add_argument(
    '--bind', dest='bind', type=str,
    help='Socket address'
)
parser.add_argument(
    '--log-level', dest='log_level', type=str,
    default='INFO'
)


def import_service(module_name):
    parts = module_name.split(":", 1)
    if len(parts) == 1:
        module_name, obj = module_name, None
    else:
        module_name, obj = parts[0], parts[1]

    try:
        __import__(module_name)
    except ImportError as e:
        if module_name.endswith(".py") and os.path.exists(module_name):
            raise exceptions.MicroServiceRunnerException(
                u"Failed to find service, did you mean '{0}'?".format(
                    module_name[:-3].replace('/', '.')
                )
            )

        raise exceptions.MicroServiceRunnerException(
            u"No module named {0}".format(
                module_name
            )
        )

    module = sys.modules[module_name]

    if not obj:
        # inspect imported module to find service
        is_micro_service = lambda x: inspect.isclass(x) and issubclass(x, BasicMicroService)
        targets = inspect.getmembers(module, predicate=is_micro_service)
        if not targets:
            raise exceptions.MicroServiceRunnerException(
                u"No service instance found in {0}".format(
                    module
                )
            )
        else:
            print targets
    else:
        try:
            cls = getattr(module, obj)
        except AttributeError:
            raise exceptions.MicroServiceRunnerException(
                u"Module {0} does not have attribute {1}".format(
                    module_name, obj
                )
            )
        else:
            return cls


def get_options():
    arguments = parser.parse_args()
    return arguments


def main():
    args = get_options()
    service = import_service(args.service)
    print service


import logging

from omnomnom import models


def validate_executable(executable):

    if isinstance(executable, models.Executable):
        if not hasattr(executable, '__call__'):
            msg = u"{0} must have __call__ defined".format(
                executable.__class__.__name__
            )
            return False, msg
        else:
            return True, None

    msg = (
        u"Cannot register executable {0}. "
        u"It must be an instance of models.Executable."
    ).format(executable)
    return False, msg


def register_executable(class_, registry, name, *args, **kwargs):

    def wrapper(c):
        instance = c(*args, **kwargs)
        is_valid, error = validate_executable(executable=instance)

        if is_valid:
            registry[name] = instance
        else:
            logging.warning(error)

        return c

    return wrapper(class_)


from setuptools import setup, find_packages

requires = [
    'nanomsg',
    'futures'
]

test_requires = [
    'mock',
    'nose',
    'coverage'
]

setup(
    name='omnomnom',
    install_requires=requires,
    tests_require=test_requires,
    packages=find_packages(),
    include_package_data=True,
    zip_safe=False,
    test_suite="nose.collector",
    version='0.1',
    description=(
        'Microservice framework with reasonable defaults'
        'using nanomsg under the hood.'
    ),
    author='jnosal',
    author_email='jacek.nosal@outlook.com',
    url='https://github.com/jnosal/omnomnom',
    keywords=[
        'nanomsg', 'microservices'
    ],
    entry_points={
        'console_scripts': [
            'omnomnom-worker = omnomnom.runners.run_worker:main',
            'omnomnom-queue = omnomnom.runners.run_queue:main',
            'omnomnom-service = omnomnom.runners.run_service:main'
        ],
    },
    classifiers=[
        'Development Status :: 1 - Planning',
        'Intended Audience :: Developers',
        'Natural Language :: English',
        'License :: OSI Approved :: MIT License',
        'Operating System :: OS Independent',
        'Programming Language :: Python',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.6',
        'Programming Language :: Python :: 2.7',
        'Topic :: Software Development :: Libraries :: Python Modules',
    ],
)



import sys
import random
import string
import logging
import importlib


def generate_name(n=4):
    return u''.join(
        random.choice(string.ascii_uppercase + string.digits)
        for _ in range(n)
    )


def get_func(name):
    module_name, attribute = name.rsplit('.', 1)
    module = importlib.import_module(module_name)
    return getattr(module, attribute)


def setup_loghandlers(level='DEBUG'):
    logger = logging.getLogger()
    logger.setLevel(level)
    formatter = logging.Formatter(u'%(asctime)s - %(levelname)s - %(message)s')
    handler = logging.StreamHandler(stream=sys.stdout)
    handler.setFormatter(formatter)
    logger.addHandler(handler)



from omnomnom.models import Executable
from omnomnom.generics.micro import Fruit


service = Fruit(
    address='tcp://127.0.0.1:5001',
    log_level='DEBUG'
)


@service.to_registry(name='ping')
class PingExecutable(Executable):

    def __call__(self, *args, **kwargs):
        return 'pong'


@service.to_registry(name='pong')
class PongExecutable(Executable):

    def __call__(self, *args, **kwargs):
        return 'ping'


@service.to_registry(name='this_wont_be_registered')
class Dummy(object):
    pass


if __name__ == '__main__':
    service.register_executable(name='ddddd', executable=PingExecutable())
    service.run()


from omnomnom.generics.micro import MicroSubClient


if __name__ == '__main__':
    address = 'tcp://127.0.0.1:5005'
    total = 0
    with MicroSubClient(address=address) as client:
        for response in client.get_responses():
            total += 1
            if total % 10000 == 0:
                print "Client received {0} messages".format(total)